<?php 

namespace WebServiceCall\Service;

use WebServiceCall\Log\OutputStreamLogger;

use WebServiceCall\Service\Interfaces\IEstabelecimento;
use WebServiceCall\Service\Estabelecimento as EstabelecimentoService;

use WebServiceCall\Repository\WS\IWebServiceHandler;
use WebServiceCall\Repository\WS\WebServiceHandler;

use WebServiceCall\Repository\DB\Procedure\ReplicaDadosCNC as ReplicaDadosProc;
use WebServiceCall\Repository\DB\Organizacao as OrgRepo;

abstract class ABaseWSService {
    private $connParams; // Estabelecimentos das quais as vendas serão buscadas.

    private $periodIni; // Período Inicial para Busca de Dados do WebService.
    private $periodEnd; // Período Final para Busca de Dados do WebService.
    
    private $serviceName; // Nome do Serviço

    private $wsHandler; // Controlador para realizar a conexão e as operações com o SOAP.
    
    private $estabelecService; // Serviço que gerencia as tratativas de estabelecimento.
    
    private $organizacaoRepo; // Repositório para obtenção de dados da Organização (NRORG).
    
    private $replicaDadosRepo; // Repositório para chamada da Procedure REPLICA_DADOS_CNC.
    
    public function __construct(array $connParams, $initialDate, $endDate, $serviceName) {
        $this->estabelecimentos = [];
        $this->connParams = $connParams;
        $this->periodIni  = $initialDate;
        $this->periodEnd  = $endDate;
        $this->serviceName  = $serviceName;
        
        $this->setWSHandler(new WebServiceHandler ());
        $this->setEstabelecimentoService(new EstabelecimentoService ()); 
    
        $this->setReplicaDadosRepo(new ReplicaDadosProc ());
        $this->setOrganizacaoRepo(new OrgRepo ());
    }

    private function setWSHandler(IWebServiceHandler $wsHandler) {
        $this->wsHandler = $wsHandler;
    }

    private function setEstabelecimentoService(IEstabelecimento $service) {
        $this->estabelecService = $service;
    }

    private function setReplicaDadosRepo($replicaDadosRepo) {
        $this->replicaDadosRepo = $replicaDadosRepo;
    }

    private function setOrganizacaoRepo($organizacaoRepo) {
        $this->organizacaoRepo = $organizacaoRepo;
    }

    private function appendEstabelecs($estabelecimentos) {
        $this->estabelecimentos = array_unique(array_merge($estabelecimentos, $this->estabelecimentos));
    }
    
    public function getInitialDate() {
        return $this->periodIni;
    }

    public function getFinalDate() {
        return $this->periodEnd;
    }
    
    public function getEstabelecimentos() {
        return $this->estabelecService->getNrEstabelecsByCodes($this->estabelecimentos);
    }

    protected function importDataFromWebService() {
        foreach($this->connParams as $connParam) {
            
            try {
                $wsParams = $this->getSoapParams($connParam);
                
                OutputStreamLogger::log(PHP_EOL . 'Buscando dados para a filial '. $connParam->getIdEmpresa());
                $wsData = $this->wsHandler->getWebServiceData($wsParams);
    
                $this->nrSeqControle = $this->wsHandler->getAccessID();
        
                $this->appendEstabelecs($this->estabelecService->checkEstabelecimentos($wsData));
                
                $this->saveWSData($wsData);
            } catch(\Exception $e) {
                OutputStreamLogger::log(PHP_EOL . 'Um erro ocorreu ao tentar importar os dados: ' . $e->getMessage() );
            }
        }
    }
    
    protected function saveWSData($data) {
        if(!empty($data) && $this->insertData($data)) {
            OutputStreamLogger::log('O relatório foi inserido com sucesso.');
            $this->replicateInsertedData();            
        } else {
            OutputStreamLogger::log('Ocorreu um erro ao tentar inserir o relatório de ' . $this->getServiceName() . '.');
        }
    }
    
    protected function replicateInsertedData() {
        if(!empty($this->nrSeqControle) && $this->replicateData()) {
            OutputStreamLogger::log('A replicação de Financeiro foi feita com sucesso.');
        } else {
            OutputStreamLogger::log('Ocorreu um erro ao tentar aplicar a replicação do Financeiro.');
        }
    }
    
    
    /*
     * Executa o sql para inserir o retorno do relatório do WebService consultado.
     *
     * @param wsArray: array com o retorno do WebService
     *
     * @return bool: booleano que sinaliza se o insert foi realizado com sucesso
     */
    private function insertData($wsArray) {
        $hasAllDataBeenSaved = true;
        
        try {
            foreach ($wsArray as $wsData) {
                $data = $this->setCommomParams($wsData);
                
                $nrEstabelec    = $data[":Estabelecimento"]; 
                $nrCtrlAcesWS   = $data[":NRCTRLACESWS"];
                $cnpj           = $data[":Cnpj"];

                // Atualiza a tabela de controle de acesso para saber conter qual o estabelecimento o WS está tratando
                $this->wsHandler->updateAccessControl($nrEstabelec);

                // Verifica se o CNPJ do relatório pertence a alguma filial do BD. 
                // Se sim, atualiza os campos de filial do estabelecimento atual
                $this->estabelecService->checkCNPJFilial($cnpj, $nrEstabelec, $data);
                $saveResult = $this->saveDataInDB($data);
                $hasAllDataBeenSaved = $hasAllDataBeenSaved && $saveResult;
            }
            
            $this->commit();
        } catch (\Exception $e) {
            $this->rollback();
            $hasAllDataBeenSaved = false;
            OutputStreamLogger::log($e->getMessage());
        }
        
        return $hasAllDataBeenSaved;
    }
    
    // Definição dos campos de auditoria
    private function setAuditFields($data) {
        $nrorg = $this->organizacaoRepo->getNRORG();
        $wsAccessDate = $this->wsHandler->getAccessDate();
        
        $data[':IDATIVO']       = 'S';
        $data[':NRORG']         = $nrorg; 
        $data[':NRORGINCLUSAO'] = $nrorg;
        $data[':NRORGULTATU']   = $nrorg;
        $data[':DTINCLUSAO']    = $wsAccessDate;
        $data[':DTULTATU']      = $wsAccessDate;
        
        return $data;
    }
    
    // Define o sequencial de controle, FK da tabela de controle de acesso
    private function setAccessControlParams($data) {
        $data[':NRCTRLACESWS'] = $this->nrSeqControle;
        return $data;
    }
    
    private function setEstabelecimento($data) {
        $data[':Estabelecimento'] = $this->estabelecService->getNrEstabelecByCode($data[':Estabelecimento']);
        return $data;
    }
    
    private function setCommomParams($data) {
        $data = $this->setAccessControlParams($data);
        $data = $this->setAuditFields($data);
        $data = $this->setEstabelecimento($data);
        
        return $this->setInsertParams($data);    
    }
    
    private function replicateData() {
        return $this->replicaDadosRepo->replicaDadosProc(
                $this->periodIni, 
                $this->periodEnd, 
                $this->nrSeqControle
        );
    }
    
    protected abstract function getSoapParams($connParam);
    
    protected abstract function setInsertParams($data);
    
    protected abstract function saveDataInDB($data);
    
    protected abstract function commit();
    
    protected abstract function rollback();
    
    private function getServiceName() {
        return $this->serviceName;
    }
}